using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.PostProcessing;
using Utils = MacGruber.Utils;

namespace CheesyFX
{
    public class CheesyDoF : MVRScript
    {
	    private static DofManager _dofDofManager;
	    private static DepthOfFieldModel model;
	    private static DepthOfFieldModel.Settings settings;
	    private ViewScan viewScan;
        public JSONStorableFloat speed = new JSONStorableFloat("Auto focus speed", 6f, 1f, 20f, false);
        public JSONStorableBool dofEnabledJSON = new JSONStorableBool("Enabled", false);
        private JSONStorableFloat aperture = new JSONStorableFloat("Aperture", 5.6f, 0.05f, 32f, false);
        private JSONStorableFloat minFocalLength = new JSONStorableFloat("Minimum focal length (at 0m)", 10f, 1f, 50f, false);
        private JSONStorableFloat maxFocalLength = new JSONStorableFloat("Maximum focal length (at 2.5m)", 100f, 1f, 300f, false);
        private JSONStorableStringChooser kernelSize = new JSONStorableStringChooser("KernelSize", Enum.GetNames(typeof(DepthOfFieldModel.KernelSize)).ToList(),"Medium", "Kernel Size");
        private JSONStorableStringChooser toggleKeyChooser = new JSONStorableStringChooser("toggleKey", Enum.GetNames(typeof(KeyCode)).ToList(), "F8", "Toggle Key");
        private static KeyCode toggleKey = KeyCode.F8;
        

        private UIDynamicToggle enabledToggle;

        private float _speed { get; set; }
        private float _minFocalLength { get; set; }
        private float _maxFocalLength { get; set; }

        private UnityAction<float> setFocusDistance;

        private bool abort;
        private JSONStorableString notInitInfo = new JSONStorableString("notInitInfo", 
	        "Press 'F8' or toggle to initialize DoF.\n\n"+
	        "Keep in mind that this will conflict with PostMagic on scene exit. Try to double tap 'm' in that case.");
        private UIDynamicButton initButton;
        private List<object> UIElements = new List<object>();

        private void OnEnable()
        {
	        abort = transform.parent.parent.name != "SessionPluginManager";
        }

        public override void Init()
        {
	        if (abort) return;
	        enabledToggle = CreateToggle(dofEnabledJSON);
	        CreateTextField(notInitInfo, false);
	        
	        // initButton = CreateButton("Init DoF", true);
	        // initButton.button.onClick.AddListener(() => dofEnabled = true);
	        enabledJSON.setCallbackFunction += val =>
	        {
		        if (!val) dofEnabledJSON.val = false;
	        };
	        dofEnabledJSON.setCallbackFunction += ToggleDof;
	        aperture.setCallbackFunction  += (float v) => { settings.aperture = v; model.settings = settings; };
	        speed.setCallbackFunction += val => _speed = val;
	        _speed = speed.val;
	        minFocalLength.setCallbackFunction += val => _minFocalLength = val;
	        _minFocalLength = minFocalLength.val;
	        maxFocalLength.setCallbackFunction += val => _maxFocalLength = val;
	        _maxFocalLength = maxFocalLength.val;
	        kernelSize.setCallbackFunction += SetKernelSize;
	        toggleKeyChooser.setCallbackFunction += SetKey;
	        
	        RegisterFloat(aperture);
	        RegisterFloat(minFocalLength);
	        RegisterFloat(maxFocalLength);
	        RegisterFloat(speed);
	        RegisterStringChooser(kernelSize);
	        RegisterStringChooser(toggleKeyChooser);
        }

        private void DeferredInit()
        {
	        RemoveTextField(notInitInfo);
	        // RemoveButton(initButton);
	        
	        _dofDofManager = new DofManager();
	        model = _dofDofManager.profile.depthOfField;
	        
	        settings.aperture = aperture.val;
	        model.settings = settings;
	        
	        // enabledToggle = CreateToggle(dofEnabledJSON);
	        // UIElements.Add(enabledToggle);
	        UIElements.Add(CreateSlider(aperture));
	        UIElements.Add(CreateSlider(minFocalLength));
	        UIElements.Add(CreateSlider(maxFocalLength));
	        UIElements.Add(CreateSlider(speed));

	        UIElements.Add(CreateScrollablePopup(kernelSize));
	        UIElements.Add(CreateScrollablePopup(toggleKeyChooser));
            
	        UIElements.Add(Utils.SetupInfoText(this, 					
	            "<b>Aperture:</b> Ratio of aperture (known as f-stop or f-number). The smaller the value is, the shallower the depth of field is.\n\n" + 
	            "<b>FocalLength:</b> Distance between the lens and the film. The larger the value is, the shallower the depth of field is.\n\n" +
	            "<b>KernelSize:</b> Convolution kernel size of the bokeh filter, which determines the maximum radius of bokeh. It also affects the performance, smaller = faster.\n\n",
	            1200.0f, true
            ));
	        SetKey(toggleKeyChooser.val);
	        
        }

        private void Deinit()
        {
	        // dofEnabled = false;
	        viewScan.uses.val--;
	        model.enabled = false;
	        _dofDofManager.Destroy();
	        _dofDofManager = null;
	        Utils.RemoveUIElements(this, UIElements);

	        notInitInfo.val = $"Press '{toggleKeyChooser.val}' or toggle to initialize DoF.\n\n" +
	                          "Keep in mind that this will conflict with PostMagic on scene exit. Try to double tap 'm' in that case.";
	        CreateTextField(notInitInfo, false);
        }
        
        private void Start()
        {
	        if (abort)
	        {
		        enabledJSON.val = false;
		        return;
	        }
	        viewScan = Utils.FindWithinSamePlugin<ViewScan>(this);
        }

        private void ToggleDof(bool val)
        {
	        if (val && _dofDofManager == null)
	        {
		        DeferredInit();
		        model.enabled = val;
		        viewScan.uses.val++;
	        }
	        else if (!val) Deinit();
        }
        
        private float GetFocalLength(float dist)
        {
	        return Mathf.Lerp(_minFocalLength, _maxFocalLength, dist*.4f);
        }

        private void OnDisable()
        {
	        ToggleDof(false);
        }

        private void OnDestroy()
        {
	        _dofDofManager.Destroy();
        }

        private void Update()
        {
	        if (Input.GetKeyDown(toggleKey)) dofEnabledJSON.val = !dofEnabledJSON.val;
	        if (!dofEnabledJSON.val) return;
	        
	        float dist = viewScan.dist.val;
	        settings.focusDistance = Mathf.Lerp(settings.focusDistance, dist, _speed * Time.deltaTime);
	        settings.focalLength = Mathf.Lerp(settings.focalLength, GetFocalLength(dist), _speed * Time.deltaTime);
	        // settings.focalLength.Print();
	        model.settings = settings;
        }
        
        private void SetKey(string val)
        {
	        toggleKey = (KeyCode)Enum.Parse(typeof(KeyCode), val);
	        enabledToggle.label = $"Enabled ({val})";
        }
        
        private void SetKernelSize(string val)
        {
	        var size = (DepthOfFieldModel.KernelSize)Enum.Parse(typeof(DepthOfFieldModel.KernelSize), val);
	        settings.kernelSize = size;
	        model.settings = settings;
        }

    }
}