using System.Collections.Generic;
using UnityEngine;
using UnityEngine.PostProcessing;
using UnityEngine.XR;
using MacGruber;

namespace CheesyFX
{
    public class DofManager
		{			
			public PostProcessingProfile profile { get; private set; }
			public Camera screenshotCamera { get; private set; }
			public Camera thumbnailCamera { get; private set; }
			public Camera windowCamera { get; private set; }
			public static bool gameIsVR { get; private set; }
					
			private PostProcessingBehaviour mainBehaviour;
			private PostProcessingBehaviour screenshotBehaviour;
			private PostProcessingBehaviour thumbnailBehaviour;
			private PostProcessingBehaviour windowBehaviour;
			private CameraHook cameraHook;
			
					
			public DofManager()
			{
				gameIsVR = XRSettings.isDeviceActive && !string.IsNullOrEmpty(XRSettings.loadedDeviceName);
				
				// Main Camera				
				Camera mainCamera = CameraTarget.centerTarget.targetCamera;
				// mainCamera.GetComponent<PostProcessingBehaviour>().NullCheck();
				mainBehaviour = Utils.GetOrAddComponent<PostProcessingBehaviour>(mainCamera);
				mainBehaviour.enabled = true;				
				profile = mainBehaviour.profile;
				if (profile == null)
					mainBehaviour.profile = profile = ScriptableObject.CreateInstance<PostProcessingProfile>();	
								
				// Main Camera - UI Separation
				cameraHook = Utils.GetOrAddComponent<CameraHook>(mainCamera);
				cameraHook.enabled = true;
				
				// Screenshot Camera
				screenshotCamera = SuperController.singleton.hiResScreenshotCamera;
				screenshotBehaviour = Utils.GetOrAddComponent<PostProcessingBehaviour>(screenshotCamera);
				screenshotBehaviour.enabled = true;
				screenshotBehaviour.profile = profile;
				
				// Thumbnail Camera
				thumbnailCamera = SuperController.singleton.screenshotCamera;
				thumbnailBehaviour = Utils.GetOrAddComponent<PostProcessingBehaviour>(thumbnailCamera);
				thumbnailBehaviour.enabled = true;
				thumbnailBehaviour.profile = profile;
				
				// Window Camera
				List<Atom> atoms = SuperController.singleton.GetAtoms();
				Atom windowAtom = atoms.Find((Atom a) => { return a.type == "WindowCamera"; });
				if (windowAtom != null)
				{
					CameraControl cameraControl = windowAtom.GetStorableByID("CameraControl") as CameraControl;
					if (cameraControl != null)
					{
						windowCamera = cameraControl.cameraToControl;
						windowBehaviour = Utils.GetOrAddComponent<PostProcessingBehaviour>(windowCamera);
						windowBehaviour.enabled = true;
						windowBehaviour.profile = profile;
					}
				}
			}					
							
			public void Destroy()
			{
				JSONStorable s;
				bool postMagicPresent = ViewScanExtensions.FindSessionPlugin("MacGruber.PostMagic.Manager", out s);
				if(!postMagicPresent) postMagicPresent = ViewScanExtensions.FindScenePlugin("PostMagic.Manager", out s);
				if(!postMagicPresent)
				{
					Object.Destroy(mainBehaviour);
					Object.Destroy(screenshotBehaviour);
					Object.Destroy(thumbnailBehaviour);
					Object.Destroy(windowBehaviour);
					Object.Destroy(cameraHook);
				}
				// else
				// {
				// 	if (mainBehaviour != null)
				// 		mainBehaviour.enabled = false;
				// 	if (screenshotBehaviour != null)
				// 		screenshotBehaviour.enabled = false;
				// 	if (thumbnailBehaviour != null)
				// 		thumbnailBehaviour.enabled = false;
				// 	if (windowBehaviour != null)
				// 		windowBehaviour.enabled = false;
				// 	if (cameraHook != null)
				// 		cameraHook.enabled = false;
				// }
			}
		}
		
		public class CameraHook : MonoBehaviour
		{
			private Camera uiCamera;
			private Camera mainCamera;
			private int uiMask = 0;
			
			private int mainCullingMask = -1;			
			
			private void OnEnable()
			{				
				uiMask = 1 << LayerMask.NameToLayer("UI")
				       | 1 << LayerMask.NameToLayer("LoadUI")
				       | 1 << LayerMask.NameToLayer("ScreenUI")
				       | 1 << LayerMask.NameToLayer("GUI");
				
				mainCamera = GetComponent<Camera>();
				mainCullingMask = mainCamera.cullingMask;
				mainCamera.cullingMask &= ~uiMask;			
				
				Transform t = transform.Find("CameraHook");
				if (t == null)
				{
					t = new GameObject("CameraHook").transform;
					t.parent = transform;
				}
				uiCamera = Utils.GetOrAddComponent<Camera>(t);
				
				uiCamera.CopyFrom(mainCamera);
				uiCamera.transform.localPosition = Vector3.zero;
				uiCamera.transform.localRotation = Quaternion.identity;
				uiCamera.cullingMask = uiMask;
				uiCamera.clearFlags = CameraClearFlags.Depth;
				uiCamera.depth = mainCamera.depth + 1;
				uiCamera.enabled = true;
			}
			
			private void OnDisable()
			{			
				if (mainCamera != null)
					mainCamera.cullingMask = mainCullingMask;
				uiCamera.enabled = false;
			}
			
			private void OnDestroy()
			{			
				if (mainCamera != null)
					mainCamera.cullingMask = mainCullingMask;
				uiCamera.enabled = false;
			}
			
			private void OnPreCull()
			{
				uiCamera.transform.localPosition = Vector3.zero;
				uiCamera.transform.localRotation = Quaternion.identity;
				if (!DofManager.gameIsVR)
					uiCamera.fieldOfView = mainCamera.fieldOfView;
			}
		}
	
}