using System;
using System.Collections;
using System.Collections.Generic;
using Leap;
using MacGruber;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Rendering;
using UnityEngine.Serialization;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace CheesyFX
{
    public class HeadLight
    {
        private MVRScript script;
        private int id;
        public string name;
        
        public GameObject lightObject;
        public Light light;
        
        private List<object> UIElements = new List<object>();
        
        public JSONStorableBool enabled;
        
        private JSONStorableStringChooser type;
        private JSONStorableStringChooser renderMode;
        public JSONStorableFloat intensity;
        
        private JSONStorableFloat spotAngle;
        private JSONStorableFloat range;
        public JSONStorableBool castShadows;
        private JSONStorableFloat shadowStrength;
        public JSONStorableColor color;

        private JSONStorableFloat customShadowResolution;
        private JSONStorableFloat shadowBias;
        private JSONStorableFloat shadowNormalBias;
        private JSONStorableFloat shadowNearPlane;
        
        private JSONStorableFloat offsetX;
        private JSONStorableFloat offsetY;
        public JSONStorableFloat offsetZ;
        public JSONStorableFloat rotation;

        private float rotationSpeed = 6f;
        
        private Button button;
        private RectTransform buttonTransform;
        private string buttonText;
        private ColorBlock activeColor;

        public BlendPreset blendPreset0;
        public BlendPreset blendPreset1;

        public BlendPreset current;
        public BlendPreset next;

        public HeadLight(MVRScript script, int id)
        {
            this.script = script;
            this.id = id;
            
            float offsetMin = 0f, offsetMax = 0f, offsetDefault = 0f, rotationDefault = 0f, intensityDefault = 0f, heightDefault = 0f, depthDefault = 0f;;
            
            switch (id)
            {
                case 0:
                    offsetMin = -3f;
                    offsetMax = 0f;
                    offsetDefault = -1f;
                    rotationDefault = 60f;
                    heightDefault = .5f;
                    intensityDefault = .5f;
                    buttonText = "L";
                    name = "Left";
                    break;
                case 1:
                    offsetMin = -1f;
                    offsetMax = 1f;
                    intensityDefault = .3f;
                    buttonText = "C";
                    name = "Center";
                    break;
                case 2:
                    offsetMin = 0f;
                    offsetMax = 3f;
                    offsetDefault = 1f;
                    heightDefault = .5f;
                    rotationDefault = -60f;
                    // rotationMin = -90f;
                    // rotationMax = 0f;
                    intensityDefault = .5f;
                    buttonText = "R";
                    name = "Right";
                    break;
                case 3:
                    offsetMin = -5f;
                    offsetMax = 5f;
                    offsetDefault = 2f;
                    heightDefault = 2.5f;
                    depthDefault = 4f;
                    rotationDefault = 180f;
                    // rotationMin = -90f;
                    // rotationMax = 0f;
                    intensityDefault = 1f;
                    buttonText = "B";
                    name = "Back";
                    break;
            }
            Vector3 offset = new Vector3(offsetDefault, heightDefault, depthDefault);
            float spotAngleDefault = LightMeUp.isOnVR ? 140f : 90f;
            
            lightObject = new GameObject();
            Transform transform = lightObject.transform;
            transform.parent = LightMeUp.cameraTransform;
            transform.localPosition = offset;
            transform.localRotation = Quaternion.identity;
            
            light = lightObject.AddComponent<Light>();

            enabled = new JSONStorableBool("Enabled", true, EnabledCallback);
            type = new JSONStorableStringChooser("Type", Enum.GetNames(typeof(LightType)).ToList(), LightType.Spot.ToString(), "Type", TypeCallback);
            renderMode = new JSONStorableStringChooser("Render Mode", Enum.GetNames(typeof(LightRenderMode)).ToList(), LightRenderMode.Auto.ToString(), "Render Mode", RenderModeCallback);
            intensity = new JSONStorableFloat("Base intensity", intensityDefault, IntensityCallback, 0f, 2f, false);
            range = new JSONStorableFloat("Range", 5f, RangeCallback, 0f, 10f, false);
            shadowStrength = new JSONStorableFloat("Shadow Strength", .2f, ShadowStrengthCallback, 0f, 1f, false);
            spotAngle = new JSONStorableFloat("Spot Angle", spotAngleDefault, SpotAngleCallback, 0f, 180f, false);
            castShadows = new JSONStorableBool("Cast Shadows", true, CastShadowsCallback);
            color = new JSONStorableColor("Color", LightMeUp.defaultColors[id].ToHSV(), OnColorChanged);
            offsetX = new JSONStorableFloat("Horizontal Offset",offset.x, OffsetCallback, offsetMin, offsetMax, false);
            offsetY = new JSONStorableFloat("Vertical Offset",offset.y, OffsetCallback, -5f, 5f, false);
            offsetZ = new JSONStorableFloat(id != 3? "Depth offset from player" : "Depth behind the target",offset.z, OffsetCallback, -5f, 5f, false);
            rotation = new JSONStorableFloat("Rotation",rotationDefault, RotationCallback, -180f, 180f, interactable:false);
            
            customShadowResolution = new JSONStorableFloat("Custom Shadow Resolution", 0f, 0f, 16f);
            shadowBias = new JSONStorableFloat("Shadow Bias (offsets casters)", .001f, 0f, 1f);
            shadowNormalBias = new JSONStorableFloat("Shadow Normal Bias (directional L)", 0f, 0f, 2f);
            shadowNearPlane = new JSONStorableFloat("Shadow near plane", .25f, 0f, 5f);

            customShadowResolution.setCallbackFunction += val => light.shadowCustomResolution = (int)val;
            shadowBias.setCallbackFunction += val => light.shadowBias = val;
            shadowNormalBias.setCallbackFunction += val => light.shadowNormalBias = val;
            shadowNearPlane.setCallbackFunction += val => light.shadowNearPlane = val;

            light.color = LightMeUp.defaultColors[id];
            range.val = 5f;
            light.intensity = intensity.defaultVal;
            light.spotAngle = spotAngle.defaultVal;
            light.type = LightType.Spot;
            lightObject.transform.localRotation = Quaternion.Euler(0f, rotationDefault, 0f);
            light.shadows = LightShadows.Hard;
            light.shadowStrength = .2f; 
            // if (id == 3) castShadows.val = false;
            light.shadowNearPlane = shadowNearPlane.val;
            light.shadowNormalBias = shadowNormalBias.val;
            light.shadowBias = shadowBias.val;
            light.shadowCustomResolution = (int) customShadowResolution.val;
            
        }
        

        public void Toggle()
        {
            enabled.val = !enabled.val;
        }

        public void EnabledCallback(bool b)
        {
            lightObject.SetActive(b);
            button.colors = b ? activeColor : LightMeUp.disabledColorBlock;
        }

        private void CastShadowsCallback(bool b)
        {
            light.shadows = b ? LightShadows.Hard : LightShadows.None;
        }

        private void ShadowStrengthCallback(float f)
        {
            light.shadowStrength = f;
            if ((id == 0 || id == 2) && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.shadowStrength = f;
                headLight.shadowStrength.val = f;
            }
        }
        

        private void TypeCallback(string x)
        {
            light.type = (LightType)Enum.Parse(typeof(LightType), x);
            if ((id == 0 || id == 2)  && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.type = (LightType)Enum.Parse(typeof(LightType), x);
                headLight.type.valNoCallback = x;
            }
            Utils.RemoveUIElements(script, UIElements);
            CreateUI();
        }

        private void RenderModeCallback(string x)
        {
            light.renderMode = (LightRenderMode)Enum.Parse(typeof(LightRenderMode), x);
            if ((id == 0 || id == 2)  && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.renderMode = (LightRenderMode)Enum.Parse(typeof(LightRenderMode), x);
                headLight.type.valNoCallback = x;
            }
        }

        private void IntensityCallback(float f)
        {
            light.intensity = f;
            if (f == 0f) enabled.val = false;
            else enabled.val = true;
            if ((id == 0 || id == 2)  && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.intensity = f;
                headLight.intensity.valNoCallback = f;
                if (f == 0f) headLight.enabled.val = false;
                else headLight.enabled.val = true;
            }
        }

        private void RangeCallback(float f)
        {
            light.range = f;
            if ((id == 0 || id == 2)  && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.range = f;
                headLight.range.valNoCallback = f;
            }
        }

        private void SpotAngleCallback(float f)
        {
            light.spotAngle = f;
            if ((id == 0 || id == 2)  && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.light.spotAngle = f;
                headLight.spotAngle.valNoCallback = f;
            }
        }

        private void OffsetCallback(float f)
        {
            lightObject.transform.localPosition = new Vector3(offsetX.val, offsetY.val, offsetZ.val);
            if ((id == 0 || id == 2) && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                headLight.lightObject.transform.localPosition = new Vector3(-offsetX.val, offsetY.val, offsetZ.val);
                headLight.offsetX.valNoCallback = -offsetX.val;
                headLight.offsetY.valNoCallback = offsetY.val;
                headLight.offsetZ.valNoCallback = offsetZ.val;
            }
        }

        private void RotationCallback(float r)
        {
            Quaternion rotation = Quaternion.Euler(0f, r, 0f);
            lightObject.transform.localRotation = rotation;
            if (id != 1 && !LightMeUp.disableSymmetry && LightMeUp.editSymmetrical.val)
            {
                HeadLight headLight = LightMeUp.headLights[id == 0 ? 2 : 0];
                rotation = Quaternion.Euler(0f, -r, 0f);
                headLight.lightObject.transform.localRotation = rotation;
                headLight.rotation.valNoCallback = -r;
            } 
        }

        public void OnColorChanged(float h, float s, float v)
        {
            Color color = Color.HSVToRGB(h, s, v);
            light.color = color;
            activeColor.normalColor = color;
            activeColor.pressedColor = color;
            activeColor.highlightedColor = color;
            if (enabled.val) button.colors = activeColor;
        }

        public void UpdateRotation(Vector3 focusPoint)
        {
            Quaternion goal = new Quaternion();
            Transform t = lightObject.transform;
            goal.SetLookRotation(focusPoint - t.position);
            t.rotation = Quaternion.Lerp(t.rotation, goal, rotationSpeed*Time.deltaTime);
            // t.rotation = goal;
            float angle = t.localRotation.eulerAngles[1];
            if (angle > 180) angle -= 360;
            rotation.valNoCallback = angle;
        }

        public void Destroy()
        {
            Object.Destroy(lightObject);
            Object.Destroy(buttonTransform.gameObject);
        }
        
        public List<object> CreateUI()
        {
            UIElements.AddRange(new object[] {
                script.CreatePopup(type),
                script.CreatePopup(renderMode),
                script.CreateSlider(intensity, false),
                script.CreateSlider(range),
                script.CreateSlider(spotAngle),
                script.CreateColorPicker(color, true),
                CreateSpacer(10f, true),
                script.CreateSlider(rotation, true),
                script.CreateSlider(offsetX, true),
                script.CreateSlider(offsetY, true),
                script.CreateSlider(offsetZ, true),
                script.CreateSlider(shadowStrength, false),
                script.CreateToggle(castShadows, false)
            });
            var textfied = script.CreateTextField(new JSONStorableString("", "Advanced shadow settings"));
            textfied.height = 40f;
            UIElements.Add(textfied);
            var slider = script.CreateSlider(customShadowResolution, false);
            slider.slider.wholeNumbers = true;
            UIElements.Add(slider);
            UIElements.Add(script.CreateSlider(shadowNearPlane, true));
            if (light.type != LightType.Directional)
            {
                slider = script.CreateSlider(shadowBias, false);
            }
            else
            {
                slider = script.CreateSlider(shadowNormalBias, false);
            }
            slider.valueFormat = "0.000";
            UIElements.Add(slider);
            return UIElements;
        }

        private UIDynamic CreateSpacer(float hight, bool rightside)
        {
            UIDynamic spacer = script.CreateSpacer(rightside);
            spacer.height = hight;
            return spacer;
        }

        public void CreateHUDToggle(RectTransform parent)
        {
            buttonTransform = script.manager.configurableScrollablePopupPrefab.transform.Find("Button") as RectTransform;
            buttonTransform = Object.Instantiate(buttonTransform, parent);
            buttonTransform.name = $"HeadLight{id}";
            // buttonTransform.anchorMax = new Vector2(.25f+.13f*(id+1)/3, 1);
            // buttonTransform.anchorMin = new Vector2(.25f+.13f*id/3, 0);
            buttonTransform.anchorMax = new Vector2(.25f+.13f*(id+1)/4, 1);
            buttonTransform.anchorMin = new Vector2(.25f+.13f*id/4, 0);
            buttonTransform.offsetMax = new Vector2(0, -5f);
            buttonTransform.offsetMin = new Vector2(0, 5f);
            button = buttonTransform.GetComponent<Button>();
            buttonTransform.Find("Text").GetComponent<Text>().text = this.buttonText;
            
            button.onClick.AddListener(Toggle);
            activeColor = ColorBlock.defaultColorBlock;
            var color1 = light.color;
            activeColor.normalColor = color1;
            activeColor.pressedColor = color1;
            activeColor.highlightedColor = color1;
            button.colors = activeColor;
        }
        
        public void BlendPresets(float val)
        {
            if (blendPreset0.spotAngle != blendPreset1.spotAngle)
                spotAngle.val = Mathf.Lerp(blendPreset0.spotAngle, blendPreset1.spotAngle, val);
            if (blendPreset0.range != blendPreset1.range)
                range.val = Mathf.Lerp(blendPreset0.range, blendPreset1.range, val);
            if (blendPreset0.shadowStrength != blendPreset1.shadowStrength)
                shadowStrength.val = Mathf.Lerp(blendPreset0.shadowStrength, blendPreset1.shadowStrength, val);
            if (blendPreset0.offsetX != blendPreset1.offsetX)
                offsetX.val = Mathf.Lerp(blendPreset0.offsetX, blendPreset1.offsetX, val);
            if (blendPreset0.offsetY != blendPreset1.offsetY)
                offsetY.val = Mathf.Lerp(blendPreset0.offsetY, blendPreset1.offsetY, val);
            if (blendPreset0.offsetZ != blendPreset1.offsetZ)
                offsetZ.val = Mathf.Lerp(blendPreset0.offsetZ, blendPreset1.offsetZ, val);
            if (blendPreset0.enabled == blendPreset1.enabled)
            {
                if (blendPreset0.enabled)
                {
                    enabled.val = true;
                    var hsv = Vector3.Lerp(blendPreset0.hsv, blendPreset1.hsv, val);
                    color.val = new HSVColor{H = hsv[0], S = hsv[1], V = hsv[2]};
                    intensity.val = Mathf.Lerp(blendPreset0.intensity, blendPreset1.intensity, val);
                }
                else
                {
                    enabled.val = false;
                    return;
                }
            }
            if (!blendPreset0.enabled)
            {
                color.val = new HSVColor{H = blendPreset1.hsv[0], S = blendPreset1.hsv[1], V = blendPreset1.hsv[2]};
                intensity.val = Mathf.Lerp(0f, blendPreset1.intensity, val);
                // if (val == 0f) enabled.val = false;
                // else
                // {
                //     // enabled.val = true;
                //     intensity.val = Mathf.Lerp(0f, blendPreset1.intensity, val);
                // }
                return;
            }
            if (!blendPreset1.enabled)
            {
                color.val = new HSVColor{H = blendPreset0.hsv[0], S = blendPreset0.hsv[1], V = blendPreset0.hsv[2]};
                intensity.val = Mathf.Lerp(blendPreset0.intensity, 0f, val);
                // if (val == 1f) enabled.val = false;
                // else
                // {
                //     // enabled.val = true;
                //     intensity.val = Mathf.Lerp(blendPreset0.intensity, 0f, val);
                // }
            }
        }
        
        public void BlendToNext(float val)
        {
            if (current.spotAngle != next.spotAngle)
                spotAngle.val = Mathf.Lerp(blendPreset0.spotAngle, next.spotAngle, val);
            if (current.range != next.range)
                range.val = Mathf.Lerp(blendPreset0.range, next.range, val);
            if (current.shadowStrength != next.shadowStrength)
                shadowStrength.val = Mathf.Lerp(current.shadowStrength, next.shadowStrength, val);
            if (current.offsetX != next.offsetX)
                offsetX.val = Mathf.Lerp(current.offsetX, next.offsetX, val);
            if (current.offsetY != next.offsetY)
                offsetY.val = Mathf.Lerp(current.offsetY, next.offsetY, val);
            if (current.offsetZ != next.offsetZ)
                offsetZ.val = Mathf.Lerp(current.offsetZ, next.offsetZ, val);
            if (current.enabled == next.enabled)
            {
                if (current.enabled)
                {
                    enabled.val = true;
                    var hsv = Vector3.Lerp(current.hsv, next.hsv, val);
                    color.val = new HSVColor{H = hsv[0], S = hsv[1], V = hsv[2]};
                    intensity.val = Mathf.Lerp(current.intensity, next.intensity, val);
                }
                else
                {
                    enabled.val = false;
                    return;
                }
            }
            if (!current.enabled)
            {
                color.val = new HSVColor{H = next.hsv[0], S = next.hsv[1], V = next.hsv[2]};
                if (val == 0f) enabled.val = false;
                else
                {
                    enabled.val = true;
                    intensity.val = Mathf.Lerp(0f, next.intensity, val);
                }
                return;
            }
            if (!next.enabled)
            {
                color.val = new HSVColor{H = current.hsv[0], S = current.hsv[1], V = current.hsv[2]};
                if (val == 1f) enabled.val = false;
                else
                {
                    enabled.val = true;
                    intensity.val = Mathf.Lerp(current.intensity, 0f, val);
                }
            }
        }

        public JSONClass GetJSON()
        {
            JSONClass jc = new JSONClass();
            enabled.Store(jc);
            color.Store(jc);
            type.Store(jc);
            renderMode.Store(jc);
            intensity.Store(jc);
            spotAngle.Store(jc);
            range.Store(jc);
            castShadows.Store(jc);
            shadowStrength.Store(jc);
            offsetX.Store(jc);
            offsetY.Store(jc);
            offsetZ.Store(jc);
            rotation.Store(jc);
            shadowBias.Store(jc);
            shadowNormalBias.Store(jc);
            shadowNearPlane.Store(jc);
            customShadowResolution.Store(jc);
            return jc;
        }

        public void LoadJSON(JSONClass jc)
        {
            bool sym = LightMeUp.editSymmetrical.val;
            LightMeUp.editSymmetrical.val = false;
            enabled.Load(jc);
            color.Load(jc);
            type.Load(jc);
            renderMode.Load(jc);
            intensity.Load(jc);
            spotAngle.Load(jc);
            range.Load(jc);
            castShadows.Load(jc);
            shadowStrength.Load(jc);
            offsetX.Load(jc);
            offsetY.Load(jc);
            offsetZ.Load(jc);
            rotation.Load(jc);
            shadowBias.Load(jc);
            shadowNormalBias.Load(jc);
            shadowNearPlane.Load(jc);
            customShadowResolution.Load(jc);
            LightMeUp.editSymmetrical.val = sym;
        }

        public void Reset()
        {
            enabled.SetValToDefault();
            color.SetValToDefault();
            type.SetValToDefault();
            renderMode.SetValToDefault();
            intensity.SetValToDefault();
            spotAngle.SetValToDefault();
            range.SetValToDefault();
            castShadows.SetValToDefault();
            shadowStrength.SetValToDefault();
            offsetX.SetValToDefault();
            offsetY.SetValToDefault();
            offsetZ.SetValToDefault();
            rotation.SetValToDefault();
            shadowBias.SetValToDefault();
            shadowNormalBias.SetValToDefault();
            shadowNearPlane.SetValToDefault();
            customShadowResolution.SetValToDefault();
        }
    }

    internal static class ColorExt
    {
        public static HSVColor ToHSV(this Color color)
        {
            float h, s, v;
            Color.RGBToHSV(color, out h, out s, out v);
            return new HSVColor
            {
                H = h,
                S = s,
                V = v
            };
        }
        
        public static Color ToRGB(this HSVColor hsv)
        {
            float h, s, v;
            return Color.HSVToRGB(hsv.H, hsv.S, hsv.V);
        }
    }
}